
-- Graham Hutton's Parsing module, cut down a bit

module Parsing where

import Char
import Monad

infixr 5 +++
newtype Parser a              =  P (String -> [(a,String)])

instance Monad Parser where
   return v                   =  P (\inp -> [(v,inp)])
   p >>= f                    =  P (\inp -> case parse p inp of
                                               []        -> []
                                               [(v,out)] -> parse (f v) out)

instance MonadPlus Parser where
   mzero                      =  P (\inp -> [])
   p `mplus` q                =  P (\inp -> case parse p inp of
                                               []        -> parse q inp
                                               [(v,out)] -> [(v,out)])
failure                       :: Parser a
failure                       =  mzero

item                          =  P (\inp -> case inp of
                                               []     -> []
                                               (x:xs) -> [(x,xs)])

parse (P p) inp               =  p inp

run (P p) inp                 =  case p inp of
                                   [(x,[])] -> x
                                   _ -> error "parsing error"

p +++ q                       =  p `mplus` q

sat p                         =  do x <- item
                                    if p x then return x else failure

digit                         =  sat isDigit
char x                        =  sat (== x)

string []                     =  return []
string (x:xs)                 =  do char x
                                    string xs
                                    return (x:xs)

many p                        =  many1 p +++ return []

many1 p                       =  do v  <- p
                                    vs <- many p
                                    return (v:vs)

nat                           :: Parser Integer
nat                           =  do xs <- many1 digit
                                    return (read xs)

int                           :: Parser Integer
int                           =  do char '-'
                                    n <- nat
                                    return (-n)
                                  +++ nat

space                         =  do many (sat isSpace)
                                    return ()

token p                       =  do space
                                    v <- p
                                    space
                                    return v

integer                       =  token int
symbol xs                     =  token (string xs)

p `chainl1` op   = do {a <- p; rest a}
                   where
                      rest a = do {f <- op; b <- p; rest (f a b)}
                               +++ return a
