
module Reduce where

import Base
import List
import Parsing


-------------------- free variables, fresh variables and substitution

fv (Var x) = [x]
fv (App n m) = fv n `union` fv m
fv (Lam x e) = x   `delete` fv e

fresh us = case ['a'..'z'] \\ us of
             (x:_)     -> x
             otherwise -> error "ran out of fresh variables!"

sub x d v@(Var y)   = if y==x then d else v
sub x d   (App n m) = App (sub x d n) (sub x d m)
sub x d a@(Lam y e) = if y==x then a else
                      if not (y `elem` fv d) then Lam y (sub x d e) else
                         let z = fresh (union [x,y] (union (fv d) (fv e)))
                         in sub x d (Lam z (sub y (Var z) e))


-------------------- one-step and multi-step reduction

red (Var _)   = Nothing
red (Lam x e) = case red e of 
                  Just e1 -> Just (Lam x e1)
                  Nothing -> Nothing
red (App (Lam x e1) e2) = Just (sub x e2 e1)
red (App e1 e2) = case red e1 of
                    Just e3 -> Just (App e3 e2)
                    Nothing -> case red e2 of
                                 Just e4 -> Just (App e1 e4)
                                 Nothing -> Nothing

reds e = case red e of
           Nothing -> []
           Just d  -> d : reds d